/*-
 * Copyright (c) 2016 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <pthread.h>
#include <stdlib.h>

#include "vcl.h"
#include "vrt.h"
#include "vas.h"
#include "vdef.h"
#include "cache/cache.h"
#include "vsb.h"
#include "miniobj.h"

#include "vcc_if.h"
#include "vre2/vre2.h"

#define ERR(ctx, msg) \
        errmsg((ctx), "vmod re2 error: " msg)

#define VERR(ctx, fmt, ...) \
        errmsg((ctx), "vmod re2 error: " fmt, __VA_ARGS__)

struct vmod_re2_regex {
	unsigned	magic;
#define VMOD_RE2_REGEX_MAGIC 0x5c3f6f24
	vre2		*vre2;
	char		*vcl_name;
	pthread_key_t	matchk;
	int		ngroups;
	unsigned	never_capture;
};

static char c;
static const void *match_failed = (void *) &c;

static size_t match_sz;
static pthread_once_t sz_init_once = PTHREAD_ONCE_INIT;

static void
errmsg(VRT_CTX, const char *fmt, ...)
{
        va_list args;

        va_start(args, fmt);
        if (ctx->method == VCL_MET_INIT) {
                AN(ctx->msg);
                VSB_vprintf(ctx->msg, fmt, args);
                VRT_handling(ctx, VCL_RET_FAIL);
        }
        else if (ctx->vsl)
                VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
        else
                /* Should this ever happen in vcl_fini() ... */
                VSL(SLT_VCL_Error, 0, fmt, args);
        va_end(args);
}

static void
init_matchsz(void)
{
	match_sz = vre2_matchsz();
}

VCL_VOID
vmod_regex__init(const struct vrt_ctx *ctx, struct vmod_re2_regex **rep,
		 const char *vcl_name, VCL_STRING pattern, VCL_BOOL utf8,
		 VCL_BOOL posix_syntax, VCL_BOOL longest_match, VCL_INT max_mem,
		 VCL_BOOL literal, VCL_BOOL never_nl, VCL_BOOL dot_nl,
		 VCL_BOOL never_capture, VCL_BOOL case_sensitive,
		 VCL_BOOL perl_classes, VCL_BOOL word_boundary,
		 VCL_BOOL one_line)
{
	struct vmod_re2_regex *re;
	const char *err;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(rep);
	AZ(*rep);
	AN(vcl_name);
	AN(pattern);
	ALLOC_OBJ(re, VMOD_RE2_REGEX_MAGIC);
	AN(re);
	*rep = re;

	AZ(pthread_key_create(&re->matchk, NULL));
	if ((err = vre2_init(&re->vre2, pattern, utf8, posix_syntax,
			     longest_match, max_mem, literal, never_nl, dot_nl,
			     never_capture, case_sensitive, perl_classes,
			     word_boundary, one_line)) != NULL) {
		VERR(ctx, "Cannot compile '%s' in %s constructor: %s", pattern,
		     vcl_name, err);
		return;
	}
	if (! never_capture) {
		if ((err = vre2_ngroups(re->vre2, &re->ngroups)) != NULL) {
			VERR(ctx, "Cannot obtain number of capturing groups in "
			     "%s constructor: %s", vcl_name, err);
			return;
		}
		assert(re->ngroups >= 0);
	}
	re->never_capture = never_capture;
	re->vcl_name = strdup(vcl_name);
	AZ(pthread_once(&sz_init_once, init_matchsz));
}

VCL_VOID
vmod_regex__fini(struct vmod_re2_regex **rep)
{
	struct vmod_re2_regex *re;

	re = *rep;
	*rep = NULL;
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);
	vre2_fini(&re->vre2);
	AZ(pthread_key_delete(re->matchk));
	if (re->vcl_name != NULL)
		free(re->vcl_name);
	FREE_OBJ(re);
}

VCL_BOOL
vmod_regex_match(const struct vrt_ctx *ctx, struct vmod_re2_regex *re,
		 VCL_STRING subject)
{
	int match = 0, ngroups = 0;
	const char *err;
	char *text = (void *) subject;
	void *group = NULL;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);
	if (subject == NULL)
		subject = "";

#define ERR_PREFIX "%s.match(\"%s\"): "

	AZ(pthread_setspecific(re->matchk, match_failed));

	if (! re->never_capture) {
		ngroups = re->ngroups + 1;
		if ((text = WS_Copy(ctx->ws, subject, -1)) == NULL) {
			VERR(ctx, ERR_PREFIX "insufficient workspace to copy "
			     "subject", re->vcl_name, subject);
			return 0;
		}
		if ((group = WS_Alloc(ctx->ws, ngroups * match_sz))
		    == NULL) {
			VERR(ctx, ERR_PREFIX "insufficient workspace to "
			     "allocate match data", re->vcl_name, subject);
			WS_Reset(ctx->ws, text);
			return 0;
		}
	}

	if ((err = vre2_match(re->vre2, text, &match, ngroups, group))
	    != NULL) {
		VERR(ctx, ERR_PREFIX "%s", re->vcl_name, subject, err);
		WS_Reset(ctx->ws, text);
		return 0;
	}

	if (! re->never_capture) {
		if (match)
			AZ(pthread_setspecific(re->matchk, group));
		else
			WS_Reset(ctx->ws, text);
	}

	return match;
#undef ERR_PREFIX
}

VCL_STRING
vmod_regex_backref(VRT_CTX, struct vmod_re2_regex *re, VCL_INT refnum,
		   VCL_STRING fallback)
{
	void *group;
	const char *err, *capture;
	char *backref;
	int len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);
	AN(fallback);
	assert(refnum >= 0);

#define ERR_PREFIX "%s.backref(%ld, \"%s\"): "

	if (re->never_capture) {
		VERR(ctx, ERR_PREFIX "never_capture is true for object %s",
		     re->vcl_name, refnum, fallback, re->vcl_name);
		return fallback;
	}
	if (refnum > re->ngroups) {
		VERR(ctx, ERR_PREFIX "backref out of range (max %d)",
		     re->vcl_name, refnum, fallback, re->ngroups);
		return fallback;
	}

	group = pthread_getspecific(re->matchk);
	if (group == NULL) {
		VERR(ctx, ERR_PREFIX "backref called without prior match",
		     re->vcl_name, refnum, fallback);
		return fallback;
	}
	if (group == match_failed)
		return fallback;

	assert((char *)group >= ctx->ws->s
	       && (char *)(group + ((re->ngroups+1) * match_sz)) <= ctx->ws->e);

	if ((err = vre2_capture(group, (int) refnum, &capture, &len))
	    != NULL) {
		VERR(ctx, ERR_PREFIX "error retrieving capture: %s",
		     re->vcl_name, refnum, fallback, err);
		return fallback;
	}

	assert(len >= 0);
	if (capture == NULL)
		return fallback;
	if (len == 0)
		return "";
	if ((backref = WS_Copy(ctx->ws, capture, len + 1)) == NULL) {
		VERR(ctx, ERR_PREFIX "insufficient workspace for backref",
		     re->vcl_name, refnum, fallback);
		return fallback;
	}
	backref[len] = '\0';
	return backref;
#undef ERR_PREFIX
}

VCL_STRING
vmod_version(const struct vrt_ctx *ctx __attribute__((unused)))
{
	return VERSION;
}
